#include <stdio.h>
#include <stdlib.h>

#if defined(WIN32)
//#  pragma comment(linker, "/subsystem:\"windows\" /entry:\"mainCRTStartup\"")
#  include "glut.h"
#  include "glext.h"
PFNGLGENFRAMEBUFFERSEXTPROC glGenFramebuffersEXT;
PFNGLBINDFRAMEBUFFEREXTPROC glBindFramebufferEXT;
PFNGLFRAMEBUFFERTEXTURE2DEXTPROC glFramebufferTexture2DEXT;
PFNGLGENRENDERBUFFERSEXTPROC glGenRenderbuffersEXT;
PFNGLBINDRENDERBUFFEREXTPROC glBindRenderbufferEXT;
PFNGLRENDERBUFFERSTORAGEEXTPROC glRenderbufferStorageEXT;
PFNGLFRAMEBUFFERRENDERBUFFEREXTPROC glFramebufferRenderbufferEXT;
PFNGLCHECKFRAMEBUFFERSTATUSEXTPROC glCheckFramebufferStatusEXT;
#elif defined(__APPLE__) || defined(MACOSX)
#  include <GLUT/glut.h>
#else
#  define GL_GLEXT_PROTOTYPES
#  include <GL/glut.h>
#endif

#include "glsl.h"

#define CHECK_FRAMEBUFFER_STATUS() \
{ \
  GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT); \
  switch(status) { \
  case GL_FRAMEBUFFER_COMPLETE_EXT: \
    break; \
  case GL_FRAMEBUFFER_UNSUPPORTED_EXT: \
    fprintf(stderr, "choose different formats\n"); \
    break; \
  default: \
    fprintf(stderr, "programming error; will fail on all hardware: %04x\n", status); \
    break; \
  } \
}

/*
** ƥ礭
*/
#define TEXWIDTH  512                                     /* ƥ */
#define TEXHEIGHT 512                                     /* ƥι⤵ */

/*
** ƥ㥪֥ȡե졼Хåե֥
*/
static GLuint tex, fb;

/*
** ֥
*/
static GLuint vertShader;
static GLuint fragShader;
static GLuint gl2Program;

/*
** 
*/
static void init(void)
{
  /* ץΥѥ롿󥯷̤ѿ */
  GLint compiled, linked;

#if defined(WIN32)
  glGenFramebuffersEXT =
    (PFNGLGENFRAMEBUFFERSEXTPROC)wglGetProcAddress("glGenFramebuffersEXT");
  glBindFramebufferEXT =
    (PFNGLBINDFRAMEBUFFEREXTPROC)wglGetProcAddress("glBindFramebufferEXT");
  glFramebufferTexture2DEXT =
    (PFNGLFRAMEBUFFERTEXTURE2DEXTPROC)wglGetProcAddress("glFramebufferTexture2DEXT");
  glGenRenderbuffersEXT =
    (PFNGLGENRENDERBUFFERSEXTPROC)wglGetProcAddress("glGenRenderbuffersEXT");
  glBindRenderbufferEXT =
    (PFNGLBINDRENDERBUFFEREXTPROC)wglGetProcAddress("glBindRenderbufferEXT");
  glRenderbufferStorageEXT =
    (PFNGLRENDERBUFFERSTORAGEEXTPROC)wglGetProcAddress("glRenderbufferStorageEXT");
  glFramebufferRenderbufferEXT =
    (PFNGLFRAMEBUFFERRENDERBUFFEREXTPROC)wglGetProcAddress("glFramebufferRenderbufferEXT");
  glCheckFramebufferStatusEXT =
    (PFNGLCHECKFRAMEBUFFERSTATUSEXTPROC)wglGetProcAddress("glCheckFramebufferStatusEXT");
#endif

  /* ƥ㥪֥Ȥ */
  glGenTextures(1, &tex);
  glBindTexture(GL_TEXTURE_2D, tex);

  /* ǥץƥγ */
  glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, TEXWIDTH, TEXHEIGHT, 0,
    GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, 0);

  /* ƥ硦̾ˡλ */
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

  /* ƥη֤ˡλ */
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

  /* ƥ㥪֥Ȥη */
  glBindTexture(GL_TEXTURE_2D, 0);

  /* ե졼Хåե֥Ȥ */
  glGenFramebuffersEXT(1, &fb);
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);

  /* ե졼Хåե֥Ȥ˥ǥץƥ */
  glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT,
    GL_TEXTURE_2D, tex, 0);

  /* 顼Хåեɤ߽񤭤ʤ */
  glDrawBuffer(GL_NONE);
  glReadBuffer(GL_NONE);

  CHECK_FRAMEBUFFER_STATUS();

  /* ե졼Хåե֥Ȥη */
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);

  /* GLSL ν */
  if (glslInit()) exit(1);

  /* ֥Ȥκ */
  vertShader = glCreateShader(GL_VERTEX_SHADER);
  fragShader = glCreateShader(GL_FRAGMENT_SHADER);

  /* Υץɤ߹ */
  if (readShaderSource(vertShader, "showdepth.vert")) exit(1);
  if (readShaderSource(fragShader, "showdepth.frag")) exit(1);

  /* СƥåΥץΥѥ */
  glCompileShader(vertShader);
  glGetShaderiv(vertShader, GL_COMPILE_STATUS, &compiled);
  printShaderInfoLog(vertShader);
  if (compiled == GL_FALSE) {
    fprintf(stderr, "Compile error in vertex shader.\n");
    exit(1);
  }

  /* ե饰ȥΥץΥѥ */
  glCompileShader(fragShader);
  glGetShaderiv(fragShader, GL_COMPILE_STATUS, &compiled);
  printShaderInfoLog(fragShader);
  if (compiled == GL_FALSE) {
    fprintf(stderr, "Compile error in fragment shader.\n");
    exit(1);
  }

  /* ץ४֥Ȥκ */
  gl2Program = glCreateProgram();

  /* ֥ȤΥץؤϿ */
  glAttachShader(gl2Program, vertShader);
  glAttachShader(gl2Program, fragShader);

  /* ֥Ȥκ */
  glDeleteShader(vertShader);
  glDeleteShader(fragShader);

  /* ץΥ */
  glLinkProgram(gl2Program);
  glGetProgramiv(gl2Program, GL_LINK_STATUS, &linked);
  printProgramInfoLog(gl2Program);
  if (linked == GL_FALSE) {
    fprintf(stderr, "Link error.\n");
    exit(1);
  }

  /* ƥ˥åȣꤹ */
  glUniform1i(glGetUniformLocation(gl2Program, "depth"), 0);
}


/****************************
** GLUT ΥХåؿ **
****************************/

/* ȥåܡѴؿ */
#include "trackball.h"

/* ؿ */
#include "scene.h"

static void display(void)
{
  GLint viewport[4];       /* ӥ塼ݡȤ¸ѡ */

  /* ߤΥӥ塼ݡȤ¸Ƥ */
  glGetIntegerv(GL_VIEWPORT, viewport);
  
  /* ӥ塼ݡȤƥΥꤹ */
  glViewport(0, 0, TEXWIDTH, TEXHEIGHT);

  /* ǥӥ塼Ѵ */
  glLoadIdentity();

  /* ΰ֤ꤹʪΤ˰ư*/
  glTranslated(0.0, 0.0, -10.0);

  /* ȥåܡ뼰βžͿ */
  glMultMatrixd(trackballRotation());

  /* ե졼Хåե֥ȤؤΥ󥰳 */
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);

  /* ǥץƥȤͭˤ */
  glEnable(GL_DEPTH_TEST);

  /* ǥץХåե򥯥ꥢ */
  glClear(GL_DEPTH_BUFFER_BIT);

  /* 褹 */
  scene();

  /* ե졼Хåե֥ȤؤΥ󥰽λ */
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

  /* ӥ塼ݡȤ򸵤᤹ */
  glViewport(viewport[0], viewport[1], viewport[2], viewport[3]);

  /* ǥץƥȤ̵ˤ */
  glDisable(GL_DEPTH_TEST);

  /* ƥ㥪֥Ȥ礹 */
  glBindTexture(GL_TEXTURE_2D, tex);

  /* ץŬѤ */
  glUseProgram(gl2Program);

  /* ̤˻ͳѷ */
  glBegin(GL_QUADS);
  glVertex2d(0.0, 0.0);
  glVertex2d(1.0, 0.0);
  glVertex2d(1.0, 1.0);
  glVertex2d(0.0, 1.0);
  glEnd();

  /* ץ */
  glUseProgram(0);

  /* ƥ㥪֥Ȥη */
  glBindTexture(GL_TEXTURE_2D, 0);

  /* ֥Хåե */
  glutSwapBuffers();
}

static void resize(int w, int h)
{
  /* ȥåܡ뤹ϰ */
  trackballRegion(w, h);

  /* ɥΤӥ塼ݡȤˤ */
  glViewport(0, 0, w, h);

  /* ƩѴ */
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(40.0, (GLdouble)w / (GLdouble)h, 5.0, 20.0);

  /* ǥӥ塼Ѵڤؤ */
  glMatrixMode(GL_MODELVIEW);
}

static void idle(void)
{
  /* ̤ؤ */
  glutPostRedisplay();
}

static void mouse(int button, int state, int x, int y)
{
  switch (button) {
  case GLUT_LEFT_BUTTON:
    switch (state) {
  case GLUT_DOWN:
    /* ȥåܡ볫 */
    trackballStart(x, y);
    glutIdleFunc(idle);
    break;
  case GLUT_UP:
    /* ȥåܡ */
    trackballStop(x, y);
    glutIdleFunc(0);
    break;
  default:
    break;
    }
    break;
  default:
    break;
  }
}

static void motion(int x, int y)
{
  /* ȥåܡư */
  trackballMotion(x, y);
}

static void keyboard(unsigned char key, int x, int y)
{
  switch (key) {
  case 'q':
  case 'Q':
  case '\033':
    /* ESC  q  Q 򥿥פ齪λ */
    exit(0);
  default:
    break;
  }
}

/*
** ᥤץ
*/
int main(int argc, char *argv[])
{
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);
  glutCreateWindow(argv[0]);
  glutDisplayFunc(display);
  glutReshapeFunc(resize);
  glutMouseFunc(mouse);
  glutMotionFunc(motion);
  glutKeyboardFunc(keyboard);
  init();
  glutMainLoop();
  return 0;
}
